package io.swagger.codegen;

import com.samskivert.mustache.Mustache;
import com.samskivert.mustache.Template;
import io.swagger.codegen.ignore.CodegenIgnoreProcessor;
import io.swagger.codegen.languages.AbstractJavaCodegen;
// NETAPP
import io.swagger.codegen.languages.PistacheServerCodegen;
import io.swagger.codegen.languages.OntapPythonClientCodegen;
// end NETAPP
import io.swagger.codegen.utils.ImplementationVersion;
import io.swagger.models.*;
import io.swagger.models.parameters.BodyParameter;
import io.swagger.models.auth.OAuth2Definition;
import io.swagger.models.auth.SecuritySchemeDefinition;
import io.swagger.models.parameters.Parameter;
import io.swagger.models.parameters.FormParameter;
import io.swagger.util.Json;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;
import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.nio.file.*;
import java.nio.charset.StandardCharsets;
import java.util.*;

public class DefaultGenerator extends AbstractGenerator implements Generator {
    protected final Logger LOGGER = LoggerFactory.getLogger(DefaultGenerator.class);
    protected CodegenConfig config;
    protected ClientOptInput opts;
    protected Swagger swagger;
// NETAPP
    protected String srcDirPrefix;
// end NETAPP
    protected CodegenIgnoreProcessor ignoreProcessor;
    protected Boolean isGenerateApis = null;
    protected Boolean isGenerateModels = null;
	protected Boolean isGenerateResources = null;
    protected Boolean isGenerateSupportingFiles = null;
    protected Boolean isGenerateApiTests = null;
    protected Boolean isGenerateApiDocumentation = null;
    protected Boolean isGenerateModelTests = null;
    protected Boolean isGenerateModelDocumentation = null;
    protected Boolean isGenerateSwaggerMetadata = true;
    protected String basePath;
    protected String basePathWithoutHost;
    protected String contextPath;
    private Map<String, String> generatorPropertyDefaults = new HashMap<>();

    @Override
    public Generator opts(ClientOptInput opts) {
        this.opts = opts;
        this.swagger = opts.getSwagger();
        this.config = opts.getConfig();
// NETAPP
        this.srcDirPrefix = opts.getSrcDirPrefix();
// end NETAPP
        this.config.additionalProperties().putAll(opts.getOpts().getProperties());

        String ignoreFileLocation = this.config.getIgnoreFilePathOverride();
        if (ignoreFileLocation != null) {
            final File ignoreFile = new File(ignoreFileLocation);
            if (ignoreFile.exists() && ignoreFile.canRead()) {
                this.ignoreProcessor = new CodegenIgnoreProcessor(ignoreFile);
            } else {
                LOGGER.warn("Ignore file specified at {} is not valid. This will fall back to an existing ignore file if present in the output directory.", ignoreFileLocation);
            }
        }

        if (this.ignoreProcessor == null) {
            this.ignoreProcessor = new CodegenIgnoreProcessor(this.config.getOutputDir());
        }

        return this;
    }

    /**
     * Programmatically disable the output of .swagger-codegen/VERSION, .swagger-codegen-ignore,
     * or other metadata files used by Swagger Codegen.
     * @param generateSwaggerMetadata true: enable outputs, false: disable outputs
     */
    @SuppressWarnings("WeakerAccess")
    public void setGenerateSwaggerMetadata(Boolean generateSwaggerMetadata) {
        this.isGenerateSwaggerMetadata = generateSwaggerMetadata;
    }

    /**
     * Set generator properties otherwise pulled from system properties.
     * Useful for running tests in parallel without relying on System.properties.
     * @param key The system property key
     * @param value The system property value
     */
    @SuppressWarnings("WeakerAccess")
    public void setGeneratorPropertyDefault(final String key, final String value) {
        this.generatorPropertyDefaults.put(key, value);
    }

    protected Boolean getGeneratorPropertyDefaultSwitch(final String key, final Boolean defaultValue) {
        String result = null;
        if (this.generatorPropertyDefaults.containsKey(key)) {
            result = this.generatorPropertyDefaults.get(key);
        }
        if (result != null) {
            return Boolean.valueOf(result);
        }
        return defaultValue;
    }

    protected String getScheme() {
        String scheme;
        if (swagger.getSchemes() != null && swagger.getSchemes().size() > 0) {
            scheme = config.escapeText(swagger.getSchemes().get(0).toValue());
        } else {
            scheme = "https";
        }
        scheme = config.escapeText(scheme);
        return scheme;
    }

    private String getHost() {
        StringBuilder hostBuilder = new StringBuilder();
        hostBuilder.append(getScheme());
        hostBuilder.append("://");
        if (!StringUtils.isEmpty(swagger.getHost())) {
            hostBuilder.append(swagger.getHost());
        } else {
            hostBuilder.append("localhost");
            LOGGER.warn("'host' not defined in the spec. Default to 'localhost'.");
        }
        if (!StringUtils.isEmpty(swagger.getBasePath()) && !swagger.getBasePath().equals("/")) {
            hostBuilder.append(swagger.getBasePath());
        }
        return hostBuilder.toString();
    }

    protected void configureGeneratorProperties() {
        // allows generating only models by specifying a CSV of models to generate, or empty for all
        // NOTE: Boolean.TRUE is required below rather than `true` because of JVM boxing constraints and type inference.
        isGenerateApis = System.getProperty(CodegenConstants.APIS) != null ? Boolean.TRUE : getGeneratorPropertyDefaultSwitch(CodegenConstants.APIS, null);
        isGenerateModels = System.getProperty(CodegenConstants.MODELS) != null ? Boolean.TRUE : getGeneratorPropertyDefaultSwitch(CodegenConstants.MODELS, null);
		isGenerateResources = System.getProperty(CodegenConstants.RESOURCES) != null ? Boolean.TRUE : getGeneratorPropertyDefaultSwitch(CodegenConstants.RESOURCES, null);
        isGenerateSupportingFiles = System.getProperty(CodegenConstants.SUPPORTING_FILES) != null ? Boolean.TRUE : getGeneratorPropertyDefaultSwitch(CodegenConstants.SUPPORTING_FILES, null);

        if (isGenerateApis == null && isGenerateModels == null && isGenerateResources == null && isGenerateSupportingFiles == null) {
            // no specifics are set, generate everything
            isGenerateApis = isGenerateModels = isGenerateResources = isGenerateSupportingFiles = true;
        } else {
            if(isGenerateApis == null) {
                isGenerateApis = false;
            }
            if(isGenerateModels == null) {
                isGenerateModels = false;
            }
			if(isGenerateResources == null) {
				isGenerateResources = false;
			}
            if(isGenerateSupportingFiles == null) {
                isGenerateSupportingFiles = false;
            }
        }
        // model/api tests and documentation options rely on parent generate options (api or model) and no other options.
        // They default to true in all scenarios and can only be marked false explicitly
        isGenerateModelTests = System.getProperty(CodegenConstants.MODEL_TESTS) != null ? Boolean.valueOf(System.getProperty(CodegenConstants.MODEL_TESTS)) : getGeneratorPropertyDefaultSwitch(CodegenConstants.MODEL_TESTS, true);
        isGenerateModelDocumentation = System.getProperty(CodegenConstants.MODEL_DOCS) != null ? Boolean.valueOf(System.getProperty(CodegenConstants.MODEL_DOCS)) : getGeneratorPropertyDefaultSwitch(CodegenConstants.MODEL_DOCS, true);
        isGenerateApiTests = System.getProperty(CodegenConstants.API_TESTS) != null ? Boolean.valueOf(System.getProperty(CodegenConstants.API_TESTS)) : getGeneratorPropertyDefaultSwitch(CodegenConstants.API_TESTS, true);
        isGenerateApiDocumentation = System.getProperty(CodegenConstants.API_DOCS) != null ? Boolean.valueOf(System.getProperty(CodegenConstants.API_DOCS)) : getGeneratorPropertyDefaultSwitch(CodegenConstants.API_DOCS, true);


        // Additional properties added for tests to exclude references in project related files
        config.additionalProperties().put(CodegenConstants.GENERATE_API_TESTS, isGenerateApiTests);
        config.additionalProperties().put(CodegenConstants.GENERATE_MODEL_TESTS, isGenerateModelTests);

        config.additionalProperties().put(CodegenConstants.GENERATE_API_DOCS, isGenerateApiDocumentation);
        config.additionalProperties().put(CodegenConstants.GENERATE_MODEL_DOCS, isGenerateModelDocumentation);

        config.additionalProperties().put(CodegenConstants.GENERATE_APIS, isGenerateApis);
        config.additionalProperties().put(CodegenConstants.GENERATE_MODELS, isGenerateModels);
		config.additionalProperties().put(CodegenConstants.GENERATE_RESOURCES, isGenerateResources);

        if(!isGenerateApiTests && !isGenerateModelTests) {
            config.additionalProperties().put(CodegenConstants.EXCLUDE_TESTS, true);
        }
        if (System.getProperty("debugSwagger") != null) {
            Json.prettyPrint(swagger);
        }
        config.processOpts();
        config.preprocessSwagger(swagger);
        config.additionalProperties().put("generatorVersion", ImplementationVersion.read());
        config.additionalProperties().put("generatedDate", DateTime.now().toString());
        config.additionalProperties().put("generatedYear", String.valueOf(DateTime.now().getYear()));
        config.additionalProperties().put("generatorClass", config.getClass().getName());
        config.additionalProperties().put("inputSpec", config.getInputSpec());
        if (swagger.getVendorExtensions() != null) {
            config.vendorExtensions().putAll(swagger.getVendorExtensions());
        }

        contextPath = config.escapeText(swagger.getBasePath() == null ? "" : swagger.getBasePath());
        basePath = config.escapeText(getHost());
        basePathWithoutHost = config.escapeText(swagger.getBasePath());

    }

    protected void configureSwaggerInfo() {
        Info info = swagger.getInfo();
        if (info == null) {
            return;
        }
        if (info.getTitle() != null) {
            config.additionalProperties().put("appName", config.escapeText(info.getTitle()));
        }
        if (info.getVersion() != null) {
            config.additionalProperties().put("appVersion", config.escapeText(info.getVersion()));
        } else {
            LOGGER.error("Missing required field info version. Default appVersion set to 1.0.0");
            config.additionalProperties().put("appVersion", "1.0.0");
        }

        if (StringUtils.isEmpty(info.getDescription())) {
            // set a default description if none if provided
            config.additionalProperties().put("appDescription",
                    "No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)");
            config.additionalProperties().put("unescapedAppDescription", "No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)");
        } else {
            config.additionalProperties().put("appDescription", config.escapeText(info.getDescription()));
            config.additionalProperties().put("unescapedAppDescription", info.getDescription());
        }

        if (info.getContact() != null) {
            Contact contact = info.getContact();
            if (contact.getEmail() != null) {
                config.additionalProperties().put("infoEmail", config.escapeText(contact.getEmail()));
            }
            if (contact.getName() != null) {
                config.additionalProperties().put("infoName", config.escapeText(contact.getName()));
            }
            if (contact.getUrl() != null) {
                config.additionalProperties().put("infoUrl", config.escapeText(contact.getUrl()));
            }
        }

        if (info.getLicense() != null) {
            License license = info.getLicense();
            if (license.getName() != null) {
                config.additionalProperties().put("licenseInfo", config.escapeText(license.getName()));
            }
            if (license.getUrl() != null) {
                config.additionalProperties().put("licenseUrl", config.escapeText(license.getUrl()));
            }
        }

        if (info.getVersion() != null) {
            config.additionalProperties().put("version", config.escapeText(info.getVersion()));
        } else {
            LOGGER.error("Missing required field info version. Default version set to 1.0.0");
            config.additionalProperties().put("version", "1.0.0");
        }

        if (info.getTermsOfService() != null) {
            config.additionalProperties().put("termsOfService", config.escapeText(info.getTermsOfService()));
        }
    }

    protected void generateModelTests(List<File> files, Map<String, Object> models, String modelName) throws IOException {
        // to generate model test files
        for (String templateName : config.modelTestTemplateFiles().keySet()) {
            String suffix = config.modelTestTemplateFiles().get(templateName);
            String filename = config.modelTestFileFolder() + File.separator + config.toModelTestFilename(modelName) + suffix;
            // do not overwrite test file that already exists
            if (new File(filename).exists()) {
                //LOGGER.info("File exists. Skipped overwriting " + filename);
                continue;
            }
            File written = processTemplateToFile(models, templateName, filename);
            if (written != null) {
                files.add(written);
            }
        }
    }

    protected void generateModelDocumentation(List<File> files, Map<String, Object> models, String modelName) throws IOException {
        for (String templateName : config.modelDocTemplateFiles().keySet()) {
            String suffix = config.modelDocTemplateFiles().get(templateName);
            String filename = config.modelDocFileFolder() + File.separator + config.toModelDocFilename(modelName) + suffix;
            if (!config.shouldOverwrite(filename)) {
                //LOGGER.info("Skipped overwriting " + filename);
                continue;
            }
            File written = processTemplateToFile(models, templateName, filename);
            if (written != null) {
                files.add(written);
            }
        }
    }

    protected void generateModels(List<File> files, List<Object> allModels, List<String> br_gen) {

        if (!isGenerateModels) {
            return;
        }

        final Map<String, Model> definitions = swagger.getDefinitions();
        if (definitions == null) {
            return;
        }

        String modelNames = System.getProperty("models");
        Set<String> modelsToGenerate = null;
        if (modelNames != null && !modelNames.isEmpty()) {
            modelsToGenerate = new HashSet<String>(Arrays.asList(modelNames.split(",")));
        }

        Set<String> modelKeys = definitions.keySet();
        LOGGER.info("Found definitions for " + modelKeys.size() + "  models");

        if (modelsToGenerate != null && !modelsToGenerate.isEmpty()) {
            Set<String> updatedKeys = new HashSet<String>();
            for (String m : modelKeys) {
                if (modelsToGenerate.contains(m)) {
                    updatedKeys.add(m);
                }
            }
            modelKeys = updatedKeys;
        }
        LOGGER.info("Updated the found definitions to now include " + modelKeys.size() + "  models");

        // store all processed models
        Map<String, Object> allProcessedModels = new TreeMap<String, Object>(new Comparator<String>() {
            @Override
            public int compare(String o1, String o2) {
                Model model1 = definitions.get(o1);
                Model model2 = definitions.get(o2);

                int model1InheritanceDepth = getInheritanceDepth(model1);
                int model2InheritanceDepth = getInheritanceDepth(model2);

                if (model1InheritanceDepth == model2InheritanceDepth) {
                    return ObjectUtils.compare(config.toModelName(o1), config.toModelName(o2));
                } else if (model1InheritanceDepth > model2InheritanceDepth) {
                    return 1;
                } else {
                    return -1;
                }
            }

            private int getInheritanceDepth(Model model) {
                int inheritanceDepth = 0;
                Model parent = getParent(model);

                while (parent != null) {
                    inheritanceDepth++;
                    parent = getParent(parent);
                }

                return inheritanceDepth;
            }

            private Model getParent(Model model) {
                if (model instanceof ComposedModel) {
                    Model parent = ((ComposedModel) model).getParent();
                    if (parent == null) {
                        // check for interfaces
                        List<RefModel> interfaces = ((ComposedModel) model).getInterfaces();
                        if (interfaces.size() > 0) {
                            RefModel interf = interfaces.get(0);
                            return definitions.get(interf.getSimpleRef());
                        }
                    }
                    if (parent != null) {
                        return definitions.get(parent.getReference());
                    }
                }

                return null;
            }
        });

        // process models only
        for (String name : modelKeys) {
            try {
                //don't generate models that have an import mapping
                if (config.importMapping().containsKey(name)) {
                    //LOGGER.info("Model " + name + " not imported due to import mapping");
                    continue;
                }
                Model model = definitions.get(name);
                Map<String, Model> modelMap = new HashMap<String, Model>();
                modelMap.put(name, model);
                Map<String, Object> models = processModels(config, modelMap, definitions);
                if (models != null) {
                    models.put("classname", config.toModelName(name));
                    models.putAll(config.additionalProperties());
                    allProcessedModels.put(name, models);
                }
            } catch (Exception e) {
                String message = "Could not process model '" + name + "'" + ". Please make sure that your schema is correct!";
                LOGGER.error(message, e);
                throw new RuntimeException(message, e);
            }
        }

        // post process all processed models
        allProcessedModels = config.postProcessAllModels(allProcessedModels);

        // generate files based on processed models
        for (String modelName : allProcessedModels.keySet()) {
            Map<String, Object> models = (Map<String, Object>) allProcessedModels.get(modelName);
            models.put("modelPackage", config.modelPackage());
            if (config instanceof OntapPythonClientCodegen) {
                models.put("pythonPackageName", config.additionalProperties().get(CodegenConstants.PACKAGE_NAME));
            }
            try {
                //don't generate models that have an import mapping
                if (config.importMapping().containsKey(modelName)) {
                    continue;
                }
                Map<String, Object> modelTemplate = (Map<String, Object>) ((List<Object>) models.get("models")).get(0);
                if (config instanceof AbstractJavaCodegen) {
                    // Special handling of aliases only applies to Java
                    if (modelTemplate != null && modelTemplate.containsKey("model")) {
                        CodegenModel m = (CodegenModel) modelTemplate.get("model");
                        if (m.isAlias) {
                            continue;  // Don't create user-defined classes for aliases
                        }
                    }
                }
                allModels.add(modelTemplate);
                for (String templateName : config.modelTemplateFiles().keySet()) {
                    String suffix = config.modelTemplateFiles().get(templateName);
                    String filename = config.modelFileFolder() + File.separator + config.toModelFilename(modelName) + suffix;
                    if (filename.endsWith(".cpp")) {
                        String br_model_path = this.srcDirPrefix + filename.replace(config.outputFolder(), "");
                        br_gen.add(br_model_path);
                    }
                    if (!config.shouldOverwrite(filename)) {
                        //LOGGER.info("Skipped overwriting " + filename);
                        continue;
                    }
                    File written = processTemplateToFile(models, templateName, filename);
                    if (written != null) {
                        files.add(written);
                    }
                }
                if(isGenerateModelTests) {
                    generateModelTests(files, models, modelName);
                }
                if(isGenerateModelDocumentation) {
                    // to generate model documentation files
                    generateModelDocumentation(files, models, modelName);
                }
            } catch (Exception e) {
                throw new RuntimeException("Could not generate model '" + modelName + "'", e);
            }
        }

        if (System.getProperty("debugModels") != null) {
            //LOGGER.info("############ Model info ############");
            Json.prettyPrint(allModels);
        }

    }

    protected void generateApis(List<File> files, List<Object> allOperations, List<Object> allModels, List<String> br_gen) {
        if (!isGenerateApis) {
            return;
        }
        Map<String, List<CodegenOperation>> paths = processPaths(swagger.getPaths());
	Set<String> apisToGenerate = null;
        String apiNames = System.getProperty("apis");
        if (apiNames != null && !apiNames.isEmpty()) {
            apisToGenerate = new HashSet<String>(Arrays.asList(apiNames.split(",")));
        }
        if (apisToGenerate != null && !apisToGenerate.isEmpty()) {
            Map<String, List<CodegenOperation>> updatedPaths = new TreeMap<String, List<CodegenOperation>>();
            for (String m : paths.keySet()) {
                if (apisToGenerate.contains(m)) {
                    updatedPaths.put(m, paths.get(m));
                }
            }
            paths = updatedPaths;
        }
        for (String tag : paths.keySet()) {
            try {
                List<CodegenOperation> ops = paths.get(tag);
                Collections.sort(ops, new Comparator<CodegenOperation>() {
                    @Override
                    public int compare(CodegenOperation one, CodegenOperation another) {
                        return ObjectUtils.compare(one.operationId, another.operationId);
                    }
                });
                Map<String, Object> operation = processOperations(config, tag, ops, allModels);

                operation.put("basePath", basePath);
                operation.put("basePathWithoutHost", basePathWithoutHost);
                operation.put("contextPath", contextPath);
                operation.put("baseName", tag);
                operation.put("apiPackage", config.apiPackage());
                operation.put("modelPackage", config.modelPackage());
                operation.putAll(config.additionalProperties());
                operation.put("classname", config.toApiName(tag));
                operation.put("classVarName", config.toApiVarName(tag));
                operation.put("importPath", config.toApiImport(tag));
                operation.put("classFilename", config.toApiFilename(tag));

                if (!config.vendorExtensions().isEmpty()) {
                    operation.put("vendorExtensions", config.vendorExtensions());
                }

                // Pass sortParamsByRequiredFlag through to the Mustache template...
                boolean sortParamsByRequiredFlag = true;
                if (this.config.additionalProperties().containsKey(CodegenConstants.SORT_PARAMS_BY_REQUIRED_FLAG)) {
                    sortParamsByRequiredFlag = Boolean.valueOf(this.config.additionalProperties().get(CodegenConstants.SORT_PARAMS_BY_REQUIRED_FLAG).toString());
                }
                operation.put("sortParamsByRequiredFlag", sortParamsByRequiredFlag);

                processMimeTypes(swagger.getConsumes(), operation, "consumes");
                processMimeTypes(swagger.getProduces(), operation, "produces");

                allOperations.add(new HashMap<String, Object>(operation));
                for (int i = 0; i < allOperations.size(); i++) {
                    Map<String, Object> oo = (Map<String, Object>) allOperations.get(i);
                    if (i < (allOperations.size() - 1)) {
                        oo.put("hasMore", "true");
                    }
                }

                for (String templateName : config.apiTemplateFiles().keySet()) {
                    String filename = config.apiFilename(templateName, tag);
                    if (filename.endsWith(".cpp")) {
                        String br_api_path = this.srcDirPrefix + filename.replace(config.outputFolder(), "");
                        br_gen.add(br_api_path);
                    }

                    if (!config.shouldOverwrite(filename) && new File(filename).exists()) {
                        //LOGGER.info("Skipped overwriting " + filename);
                        continue;
                    }

                    File written = processTemplateToFile(operation, templateName, filename);
                    if (written != null) {
                        files.add(written);
                    }
                }

                if(isGenerateApiTests) {
                    // to generate api test files
                    for (String templateName : config.apiTestTemplateFiles().keySet()) {
                        String filename = config.apiTestFilename(templateName, tag);
                        // do not overwrite test file that already exists
                        if (new File(filename).exists()) {
                            //LOGGER.info("File exists. Skipped overwriting " + filename);
                            continue;
                        }

                        File written = processTemplateToFile(operation, templateName, filename);
                        if (written != null) {
                            files.add(written);
                        }
                    }
                }


                if(isGenerateApiDocumentation) {
                    // to generate api documentation files
                    for (String templateName : config.apiDocTemplateFiles().keySet()) {
                        String filename = config.apiDocFilename(templateName, tag);
                        if (!config.shouldOverwrite(filename) && new File(filename).exists()) {
                            //LOGGER.info("Skipped overwriting " + filename);
                            continue;
                        }

                        File written = processTemplateToFile(operation, templateName, filename);
                        if (written != null) {
                            files.add(written);
                        }
                    }
                }

            } catch (Exception e) {
                throw new RuntimeException("Could not generate api file for '" + tag + "'", e);
            }
        }
        if (System.getProperty("debugOperations") != null) {
            //LOGGER.info("############ Operation info ############");
            Json.prettyPrint(allOperations);
        }

    }

    private Map<String, Object> parseModels() {
        final Map<String, Model> definitions = swagger.getDefinitions();
        if (definitions == null) {
            return null;
        }

        String modelNames = System.getProperty("models");
        Set<String> modelsToGenerate = null;
        if (modelNames != null && !modelNames.isEmpty()) {
            modelsToGenerate = new HashSet<String>(Arrays.asList(modelNames.split(",")));
        }

        Set<String> modelKeys = definitions.keySet();
        if (modelsToGenerate != null && !modelsToGenerate.isEmpty()) {
            Set<String> updatedKeys = new HashSet<String>();
            for (String m : modelKeys) {
                if (modelsToGenerate.contains(m)) {
                    updatedKeys.add(m);
                }
            }
            modelKeys = updatedKeys;
        }

        // store all processed models
        Map<String, Object> allProcessedModels = new TreeMap<String, Object>(new Comparator<String>() {
            @Override
            public int compare(String o1, String o2) {
                Model model1 = definitions.get(o1);
                Model model2 = definitions.get(o2);

                int model1InheritanceDepth = getInheritanceDepth(model1);
                int model2InheritanceDepth = getInheritanceDepth(model2);

                if (model1InheritanceDepth == model2InheritanceDepth) {
                    return ObjectUtils.compare(config.toModelName(o1), config.toModelName(o2));
                } else if (model1InheritanceDepth > model2InheritanceDepth) {
                    return 1;
                } else {
                    return -1;
                }
            }

            private int getInheritanceDepth(Model model) {
                int inheritanceDepth = 0;
                Model parent = getParent(model);

                while (parent != null) {
                    inheritanceDepth++;
                    parent = getParent(parent);
                }

                return inheritanceDepth;
            }

            private Model getParent(Model model) {
                if (model instanceof ComposedModel) {
                    Model parent = ((ComposedModel) model).getParent();
                    if (parent == null) {
                        // check for interfaces
                        List<RefModel> interfaces = ((ComposedModel) model).getInterfaces();
                        if (interfaces.size() > 0) {
                            RefModel interf = interfaces.get(0);
                            return definitions.get(interf.getSimpleRef());
                        }
                    }
                    if (parent != null) {
                        return definitions.get(parent.getReference());
                    }
                }

                return null;
            }
        });

        // process models only
        for (String name : modelKeys) {
            try {
                //don't generate models that have an import mapping
                if (config.importMapping().containsKey(name)) {
                    //LOGGER.info("Model " + name + " not imported due to import mapping");
                    continue;
                }
                Model model = definitions.get(name);
                Map<String, Model> modelMap = new HashMap<String, Model>();
                modelMap.put(name, model);
                Map<String, Object> models = processModels(config, modelMap, definitions);
                if (models != null) {
                    models.put("classname", config.toModelName(name));
                    models.putAll(config.additionalProperties());
                    allProcessedModels.put(name, models);
                }
            } catch (Exception e) {
                String message = "Could not process model '" + name + "'" + ". Please make sure that your schema is correct!";
                LOGGER.error(message, e);
                throw new RuntimeException(message, e);
            }
        }

        // post process all processed models
        allProcessedModels = config.postProcessAllModels(allProcessedModels);
        return allProcessedModels;
    }

    /**
       Marshmallow types start with a capital letter.
    */
    private void setMarshmallowType(CodegenProperty var) {
        setMarshmallowType(var, false);
    }

    private void setMarshmallowType(CodegenProperty var, boolean isList) {
        if (var.isString)  {
            var.marshmallowType = "fields.Str";
            if (isList) {
                var.vendorExtensions.put("cli_type", "List[str]");
            } else {
                var.vendorExtensions.put("cli_type", "str");
            }
        } else if (var.isInteger || var.isLong) {
            var.marshmallowType = "Size";
            if (isList) {
                var.vendorExtensions.put("cli_type", "List[Size]");
            } else {
                var.vendorExtensions.put("cli_type", "Size");
            }
        } else if (var.isDateTime) {
            var.marshmallowType = "ImpreciseDateTime";
            if (isList) {
                var.vendorExtensions.put("cli_type", "List[datetime]");
            } else {
                var.vendorExtensions.put("cli_type", "datetime");
            }
        } else if (var.isBoolean) {
            var.marshmallowType = "fields.Boolean";
            if (isList) {
                var.vendorExtensions.put("cli_type", "List[bool]");
            } else {
                var.vendorExtensions.put("cli_type", "bool");
            }
        } else if (var.isNumber) {
            var.marshmallowType = "fields.Number";
            if (isList) {
                var.vendorExtensions.put("cli_type", "List[float]");
            } else {
                var.vendorExtensions.put("cli_type", "float");
            }
        }
    }

    /**
       Turn /api/path/{object.field.subfield} into /api/path/{object[field][subfield]}
    */
    private String toPythonDict(String url) {
        StringBuilder pyDictURL = new StringBuilder();
        for (String urlPiece : url.split("/")) {
            if (urlPiece.startsWith("{")) {
                String[] items = urlPiece.substring(1, urlPiece.length() - 1).split("\\.");
                pyDictURL.append("/{" + items[0]);
                for (int i = 1; i < items.length; i++) {
                    pyDictURL.append("[");
                    pyDictURL.append(items[i]);
                    pyDictURL.append("]");
                }
                pyDictURL.append("}");
            } else if (urlPiece.length() > 0) {
                if (!urlPiece.startsWith("/")) {
                    urlPiece = "/" + urlPiece;
                }
                pyDictURL.append(urlPiece);
            }
        }
        return pyDictURL.toString();
    }

    private Map<String, Object> linkPathsAndModels() {
        // Parse the models from YAML POJOs to POJOs w/ mustache variables
        Map<String, Object> modelsWithProps = parseModels();
        Map<String, Object> models = new HashMap<String, Object>();
        Map<String, Object> modelRefs = new HashMap<String, Object>();
        Map<String, Object> fields = new HashMap<String, Object>();
        for (String modelName : modelsWithProps.keySet()) {
            // only add to list of generated sources if it is a model that isn't
            // 1. the result of INLINE_REF() getting resolved (they're removed in InlineModelResolver)
            // 2. a _reference to a different model
            // 3. a _response that is an array of models
            // all this together makes sure we only have 1 model generated
            // corresponding to each object in ONTAP's REST object mdoeling
            String strippedName = modelName;
            if (modelName.endsWith("_reference")) {
                strippedName = modelName.substring(0, modelName.length() - 10);
            } else if (modelName.endsWith("_response")) {
                strippedName = modelName.substring(0, modelName.length() - 9);
            }
            boolean doAdd = false;
            boolean doRefAdd = false;
            if (!modelName.endsWith("_reference") &&
                !modelName.endsWith("_response") &&
                !modelName.endsWith("_instance") &&
                !modelName.startsWith("inline_response")) {
                doAdd = true;
            }  else if (modelName.endsWith("_reference")) {
                doRefAdd = true;
                // if the model ends with _reference, but there is no corresponding
                // non-reference model, then lets generate this one. This is really
                // a problem with the yaml that should be fixed at that layer, but
                // we don't have time for all that
                if (modelsWithProps.get(strippedName) == null) {
                    doAdd = true;
                }
            } else if (modelName.endsWith("_response")) {
                if (modelsWithProps.get(strippedName) == null) {
                    doAdd = true;
                }
            }

            HashMap<String, Object> modelAndProps = (HashMap<String, Object>)modelsWithProps.get(modelName);
            ArrayList<HashMap> modelList = (ArrayList<HashMap>)modelAndProps.get("models");
            HashMap<String, Object> modelMap = (HashMap<String, Object>) modelList.get(0);
            CodegenModel cm = (CodegenModel)modelMap.get("model");

            if (doAdd) {
                if (cm.hasVars) {
                    cm.classname = config.toModelName(strippedName);
                    models.put(strippedName, modelList.get(0));
                } else {
                    fields.put(cm.classname, cm);
                }
            }

            if (doRefAdd) {
                modelRefs.put(modelName, modelList.get(0));
            }
        }

        // Adjust the resources schema
        for (String modelName : models.keySet()) {
            HashMap<String, Object> schema = (HashMap<String, Object>)models.get(modelName);
            CodegenModel model = (CodegenModel)schema.get("model");
            model.name = config.toResourceFilename(modelName);
            schema.put("cli_name", model.name.replaceAll("_", " "));

            if (model.unescapedDescription != null) {
                schema.put("pdocModelDescription", markdownAdjust(model.unescapedDescription));
            }

            for (CodegenProperty var : model.vars) {
                var.name = config.toParamName(var.name);
                var.vendorExtensions.put("cli_name", var.name);
                if (var.isPrimitiveType) {
                    setMarshmallowType(var);
                    if (var.isListContainer) {
                        setMarshmallowType(var.items, true);
                    }
                }
                if (var.complexType != null) {
                    // Some types defined as models, but are really just wrappers
                    // over primitive types. Turn those back into primitive types
                    CodegenModel varModel = (CodegenModel)fields.get(var.complexType);
                    if (varModel != null) {
                        // currently the only primitive type that is wrapped by
                        // a model is a string
                        var.isPrimitiveType = true;
                        var.marshmallowType = "fields.Str";
                        var.vendorExtensions.put("cli_type", "str");
                        var.allowableValues = varModel.allowableValues; // copy validation rules
                        if (varModel.isArrayModel) {
                            var.isListContainer = true;
                        }
                        if (var.isListContainer && var.items != null) {
                            var.items.marshmallowType = "fields.Str";
                            var.vendorExtensions.put("cli_type", "List[str]");
                        }
                    }

                    // point all *_reference, *_response, and *_instance objects back to their original object
                    if (var.complexType.endsWith("Reference")) {
                        var.complexType = var.complexType.substring(0, var.complexType.length() - 9);
                    } else if (var.complexType.endsWith("Response")) {
                        var.complexType = var.complexType.substring(0, var.complexType.length() - 8);
                    }
                }
                String desc = var.unescapedDescription;
                if (desc == null) {
                    desc = "The " + var.name + " field of the " + model.name + ".";
                }
                if (var.isEnum && var._enum != null) {
                    desc += "\n\nValid choices:\n\n";
                    for (String choice : var._enum) {
                        desc += "* " + choice + "\n";
                    }
                }
                if (!var.isEnum && var.example != null) {
                    desc += "\n\nExample: " + var.example;
                }
                var.vendorExtensions.put("pdocFieldDescription", markdownAdjust(desc));
            }

            String[] nonPatchable = {"x-ntap-readOnly", "x-ntap-readCreate", "x-ntap-createOnly"};
            schema.put("patchable_fields", this.getAccessible(model, models, modelRefs, nonPatchable));

            String[] nonPostable = {"x-ntap-readOnly", "x-ntap-readModify", "x-ntap-modifyOnly"};
            schema.put("postable_fields", this.getAccessible(model, models, modelRefs, nonPostable));

            String[] nonGettable = {"x-ntap-createOnly", "x-ntap-modifyOnly", "x-ntap-writeOnly"};
            schema.put("gettable_fields", this.getAccessible(model, models, modelRefs, nonGettable));
        }

        /* For each path:
           1. Get it's keys, location
           2. Look up what model it's associated with
           3. Add keys, location, ops mustache vars to that model's mustache vars
           4. Add custom POST actions
           5. Fix up path key names (e.g. /roles/{name}/privileges => /roles/{role[name]}/privileges)
        */
        Map<String, io.swagger.models.Path> paths = this.swagger.getPaths();
        for (String url : paths.keySet()) {
            // 1.
            io.swagger.models.Path path = paths.get(url);
            String[] urlPieces = url.split("/");
            ArrayList<HashMap<String, Object>> keys = new ArrayList<>();
            ArrayList<HashMap<String, Object>> nativeKeys = new ArrayList<>();
            ArrayList<HashMap<String, Object>> parentKeys = new ArrayList<>();
            ArrayList<HashMap<String, Object>> postFormDataArray = new ArrayList<>();
            ArrayList<HashMap<String, Object>> patchFormDataArray = new ArrayList<>();
            int lastChildLoc = -1;
            for (int i = urlPieces.length - 1; i >= 0; i--) {
                if (urlPieces[i].startsWith("{")) {
                    HashMap<String, Object> keyEntry = new HashMap<>();
                    keyEntry.put("key", urlPieces[i].substring(1, urlPieces[i].length() - 1));
                    keys.add(keyEntry);
                    if (lastChildLoc < 0) {
                        nativeKeys.add(keyEntry);
                    } else {
                        parentKeys.add(keyEntry);
                    }
                } else if (lastChildLoc < 0) {
                    lastChildLoc = i;
                }
            }

            Collections.reverse(keys);
            if (keys.size() > 0) {
                keys.get(keys.size() - 1).put("lastKey", true);
            }
            // Wish we had String.join() in Java 1.7...
            StringBuilder urlNoChildKeys = new StringBuilder();
            if (lastChildLoc > 0) {
                for (int j = 1; j <= lastChildLoc; j++) {
                    urlNoChildKeys.append("/" + urlPieces[j]);
                }
            } else {
                urlNoChildKeys.append(url);
            }

            // 2.
            String ref = null;
            String ref_type = null;
            boolean isSingleton = false;
            boolean isFormPost = false;
            if (path.getPost() != null) {
                for (Parameter param : path.getPost().getParameters()) {
                    if (param.getIn().equals("body")) {
                        BodyParameter bodyParam = (BodyParameter)param;
                        ref = bodyParam.getSchema().getReference();
                    }
                    if (param.getIn().equals("formData")) {
                        isFormPost = true;
                        HashMap<String, Object> postFormDataEntry = new HashMap<>();
                        FormParameter formParam = (FormParameter)param;
                        ref = formParam.getName();
                        ref_type = formParam.getType();
                        postFormDataEntry.put("paramName",ref);
                        postFormDataEntry.put("paramType",ref_type);
                        postFormDataArray.add(postFormDataEntry);
                    }
                }
            }

            String patch_ref = null;
            String patch_ref_type = null;
            if (path.getPatch() != null){
                for (Parameter param : path.getPatch().getParameters()){
                    if (param.getIn().equals("formData")){
                        HashMap<String, Object> patchFormDataEntry = new HashMap<>();
                        FormParameter formParam = (FormParameter)param;
                        patch_ref = formParam.getName();
                        patch_ref_type = formParam.getType();
                        patchFormDataEntry.put("paramName",patch_ref);
                        patchFormDataEntry.put("paramType",patch_ref_type);
                        patchFormDataArray.add(patchFormDataEntry);
                    }
                }
            }

            boolean hasNativeKeysAndRecords = false;
            if (path.getGet() != null) {
                String getRef = null;
                Map<String, Response> resps = path.getGet().getResponses();
                for (String httpCode : resps.keySet()) {
                    if (httpCode.equals("200")) {
                        getRef = resps.get(httpCode).getResponseSchema().getReference();
                        if (getRef != null && !nativeKeys.isEmpty() && getRef.endsWith("_response")) {
                            // There can be an edge case where a path is not a singleton but will have
                            // records, num_records and will have native keys. Figure that out here to
                            // handle it
                            String respRef = getRef;
                            if (respRef.startsWith("#/definitions/")) {
                                respRef = respRef.substring(14);
                            }

                            HashMap<String, Object> modelAndProps = (HashMap<String, Object>)modelsWithProps.get(respRef);
                            ArrayList<HashMap> modelList = (ArrayList<HashMap>)modelAndProps.get("models");
                            HashMap<String, Object> modelMap = (HashMap<String, Object>) modelList.get(0);
                            CodegenModel cgModel = (CodegenModel)modelMap.get("model");
                            hasNativeKeysAndRecords = hasRecordsAndNumRecords(cgModel);

                            // some instance gets have their objects misnamed with
                            // _response in the ONTAP yaml.
                            getRef = getRef.substring(0, getRef.length() - 9);
                        }
                    }
                }
                if (!nativeKeys.isEmpty() || path.getPost() == null || isFormPost) {
                    ref = getRef;
                }

                HashMap<String, Object> model = null;
                if (getRef != null) {
                    if (getRef.startsWith("#/definitions/")) {
                        getRef = getRef.substring(14);
                    }
                    if (getRef.endsWith("_reference")) {
                        getRef = getRef.substring(0, getRef.length() - 10);
                    }
                    if (getRef.endsWith("_instance")) {
                        getRef = getRef.substring(0, getRef.length() - 9);
                    }
                    model = (HashMap<String, Object>)models.get(getRef);
                }
                if (model != null && nativeKeys.isEmpty()) {
                    CodegenModel cgModel = (CodegenModel)model.get("model");
                    if (!hasRecordsAndNumRecords(cgModel)) {
                        //LOGGER.info("GET " + url + " looks like a singleton");
                        isSingleton = true;
			if (ref == null) {
			    // must be a singleton without a post body
			    ref = getRef;
			}
                    }
                }
            }

            // There's an edge case where we can have post, get_collection, and delete
            // but delete has no model, so we end up missing the delete. Fix that here.
            if (ref == null && path.getDelete() != null) {
                io.swagger.models.Path collection_path = paths.get(urlNoChildKeys.toString());
                if (collection_path != null) {
                    if (collection_path.getPost() != null) {
                        for (Parameter param : collection_path.getPost().getParameters()) {
                            if (param.getIn().equals("body")) {
                                BodyParameter bodyParam = (BodyParameter)param;
                                ref = bodyParam.getSchema().getReference();
                                break;
                            }
                        }
                    } else if (collection_path.getGet() != null) {
                        Map<String, Response> resps = collection_path.getGet().getResponses();
                        for (String httpCode : resps.keySet()) {
                            if (httpCode.equals("200")) {
                                ref = resps.get(httpCode).getResponseSchema().getReference();

                                // some instance gets have their objects misnamed with
                                // _response in the ONTAP yaml. All of the instance
                                // gets in the Deploy yaml are purposfully this way.
                                // This adjustment should fix both.
                                if (ref != null && !nativeKeys.isEmpty() && ref.endsWith("_response")) {
                                    ref = ref.substring(0, ref.length() - 9);
                                }
                            }
                        }
                    }
                }
            }

            // 3.
            if (ref != null) {
                if (ref.startsWith("#/definitions/")) {
                    ref = ref.substring(14);
                }
                if (ref.endsWith("_reference")) {
                    ref = ref.substring(0, ref.length() - 10);
                }
                if (ref.endsWith("_response")) {
                    ref = ref.substring(0, ref.length() - 9);
                }
                if (ref.endsWith("_instance")) {
                    ref = ref.substring(0, ref.length() - 9);
                }

                HashMap<String, Object> model = (HashMap<String, Object>)models.get(ref);
                if (model != null) {
                    if (keys.size() > 0) {
                        if (!model.containsKey("keys")) {
                            model.put("keys", keys);
                            model.put("hasKeys", true);
                            if (parentKeys.size() > 0) {
                                ArrayList<String> cli_keys = new ArrayList<String>();
                                for (HashMap<String, Object> keyEntry : parentKeys) {
                                    String keyName = (String)keyEntry.get("key");
                                    cli_keys.add(keyName.replace(".", "_"));
                                }
                                model.put("cli_keys", cli_keys);

                                model.put("vars_without_cli_keys", ((CodegenModel)model.get("model")).vars);
                                for (String cli_key : cli_keys) {
                                    for (Iterator<CodegenProperty> iterator =
                                            ((List<CodegenProperty>)model.get("vars_without_cli_keys")).iterator();
                                        iterator.hasNext(); ) {
                                        if (iterator.next().baseName.equals(cli_key)) {
                                            iterator.remove();
                                        }
                                    }
                                }
                            }
                        } else {
                            // This happens in the case where a collection endpoint
                            // that has parent keys is processed before the instance
                            // version of that endpoint. Since we have all of the
                            // keys in the instance version, just write over it.
                            ArrayList<HashMap<String, Object>> existingKeys = (ArrayList<HashMap<String, Object>>)model.get("keys");
                            if (keys.size() > existingKeys.size()) {
                                model.put("keys", keys);
                            }
                        }
                    }

                    if (postFormDataArray.size() > 0){
                        model.put("postFormDataArray", postFormDataArray);
                        model.put("postFormData", true);
                    }

                    if (patchFormDataArray.size() > 0){
                        model.put("patchFormDataArray", patchFormDataArray);
                        model.put("patchFormData", true);
                    }

                    if (!model.containsKey("url")) {
                        model.put("url", toPythonDict(urlNoChildKeys.toString()));
                    }
                    if (!model.containsKey("basePath")) {
                        if (this.basePathWithoutHost != null) {
                            model.put("basePath", this.basePathWithoutHost);
                        } else {
                            model.put("basePath", this.basePath);
                        }
                    }

                    HashMap<String, Object> longDesc = (HashMap<String, Object>)path.getVendorExtensions().get("x-ntap-long-description");
                    if (longDesc != null && longDesc.get("description") != null) {
                        String desc = (String)longDesc.get("description");
                        model.put("pdocLongDescription", markdownAdjust(desc));
                    }

                    HashMap<String, Object> ops = (HashMap<String, Object>)model.get("operations");
                    if (ops == null) {
                        ops = new HashMap<>();
                    }
                    if (path.getPost() != null) {
                        ops.put("post", path.getPost().getDescription());
                        if (isBulkPostSupported(models, ref, path) && !isSingleton) {
                            ops.put("post_collection", path.getPost().getDescription());
                        }
                    }
                    if (path.getPatch() != null) {
                        if (!isSingleton && nativeKeys.isEmpty()) {
                            ops.put("patch_collection", path.getPatch().getDescription());
                        } else {
                            ops.put("patch", path.getPatch().getDescription());
                            if (!isSingleton) {
                                ops.put("patch_collection", path.getPatch().getDescription());
                            }
                        }
                    }
                    if (path.getDelete() != null) {
                        if (!isSingleton && nativeKeys.isEmpty()) {
                            ops.put("delete_collection", path.getDelete().getDescription());
                        } else {
                            ops.put("delete", path.getDelete().getDescription());
                            if (!isSingleton) {
                                ops.put("delete_collection", path.getDelete().getDescription());
                            }
                        }
                    }
                    if (path.getGet() != null) {
                        if (!isSingleton && nativeKeys.isEmpty()) {
                            ops.put("get_collection", path.getGet().getDescription());
                            ops.put("find", path.getGet().getDescription());
                            ops.put("count_collection", path.getGet().getDescription());
                        } else {
                            ops.put("get", path.getGet().getDescription());
                            if (!isSingleton && hasNativeKeysAndRecords) {
                                ops.put("get_collection", path.getGet().getDescription());
                                ops.put("find", path.getGet().getDescription());
                                ops.put("count_collection", path.getGet().getDescription());
                            }
                        }
                    }
                    model.put("operations", ops);
                }
            } else if (ref != null && url.startsWith("/")) {
                LOGGER.warn("While looking at " + url + ", no parsed model was found for " + ref + ".");
            }
        }

        // 4.
        ArrayList<HashMap<String, Object>> actionFormDataArray = new ArrayList<>();
        for (String url : paths.keySet()) {
            io.swagger.models.Path path = paths.get(url);

            if (path.getPost() == null || path.getGet() != null ||
                path.getPatch() != null || path.getDelete() != null) {
                continue;
            }

            ArrayList<String> postOnlyPaths = new ArrayList<String>() {
                {
                    add("/storage/snaplock/litigations/{litigation.id}/operations");
                    add("/protocols/file-security/permissions/{svm.uuid}/{path}/acl");
                    add("/name-services/unix-groups/{svm.uuid}/{unix_group.name}/users");
                    add("/support/ems/application-logs");
                }
            };
            if (postOnlyPaths.contains(url)) {
                // This is a special case. It really isn't an action, but it only
                // has a POST on the collection which is odd so we'll just exempt it
                continue;
            }

            // since there's only a POST an no other verb, then this is a
            // custom verb on a resource. Find it and add this verb.

            String action = url.substring(url.lastIndexOf('/') + 1);
            url = url.substring(0, url.lastIndexOf('/'));
            String[] urlPieces = url.split("/");
            int lastChildLoc = -1;
            for (int i = urlPieces.length - 1; i >= 0; i--) {
                if (!urlPieces[i].startsWith("{") && lastChildLoc <0) {
                    lastChildLoc = i;
                }
            }
            // Wish we had String.join() in Java 1.7...
            StringBuilder urlNoChildKeys = new StringBuilder();
            String base = "";
            if (lastChildLoc > 0) {
                for (int j = 0; j <= lastChildLoc; j++) {
                    urlNoChildKeys.append("/" + urlPieces[j]);
                }
                base = urlNoChildKeys.toString().substring(1);
            } else {
                urlNoChildKeys.append(url);
                base = urlNoChildKeys.toString().substring(1);
            }

            // try to find a resource to pin this action on. We're looking for an
            // object that has the same base URL as this action.
            // For example, if the action was "sign" and the base (without keys)
            // was /security/certificates, then we will attach the sign action
            // to the security_certificate model
            for (Object value : models.values()) {
                HashMap<String, Object> model = (HashMap<String, Object>)value;
                String modelUrl = (String)model.get("url");
                CodegenModel codegenModel = (CodegenModel)model.get("model");
                if (modelUrl != null) {
                    modelUrl = modelUrl.replaceAll("\\[", ".").replaceAll("\\]", "");
                    LOGGER.debug(
                        "Base is '" + base + "'. Verb is '" + action + "'. URL for " +
                        codegenModel.name + " is '" + modelUrl + "'."
                    );
                    if (modelUrl.equals(base)) {
                        ArrayList<HashMap<String, String>> actions = (ArrayList<HashMap<String, String>>)model.get("post_actions");
                        if (actions == null) {
                            actions = new ArrayList<HashMap<String, String>>();
                        }

                        String ref = null;
                        String ref_type = null;
                        if (path.getPost() != null) {
                            for (Parameter param : path.getPost().getParameters()) {
                                if (param.getIn().equals("formData")) {
                                    HashMap<String, Object> actionFormDataEntry = new HashMap<>();
                                    FormParameter formParam = (FormParameter)param;
                                    ref = formParam.getName();
                                    ref_type = formParam.getType();
                                    actionFormDataEntry.put("paramName",ref);
                                    actionFormDataEntry.put("paramType",ref_type);
                                    actionFormDataArray.add(actionFormDataEntry);
                                }
                            }
                        }

                        HashMap<String, String> actionMap = new HashMap<String, String>();
                        String actionName = action.replaceAll("-", "_");
                        for (CodegenProperty var : codegenModel.vars) {
                            // check for a case where the name of the action matches
                            // the name of a field in the model. For example, the
                            // /api/snapmirror/relationships/{uuid}/restore endpoint
                            // does this
                            if (var.name.equals(actionName)) {
                                actionName += "_action";
                            }
                        }
                        actionMap.put("actionName", actionName);
                        actionMap.put("actionPath", action);
                        actionMap.put("description", path.getPost().getDescription());
                        actions.add(actionMap);
                        model.put("post_actions", actions);
                        if (actionFormDataArray.size() > 0){
                            model.put("actionFormDataArray", actionFormDataArray);
                            model.put("actionFormData", true);
                        }

                        // The action object is not useful now, so we remove it
                        // by taking away the actions we assigned to it which causes
                        // us to not output the file later
                        for (Object actionValue : models.values()) {
                            HashMap<String, Object> actionModel = (HashMap<String, Object>)actionValue;
                            String actionModelUrl = (String)actionModel.get("url");
                            if (actionModelUrl == null) {
                                continue;
                            }
                            if (actionModelUrl.startsWith(base) && actionModelUrl.endsWith(action)) {
                                LOGGER.info("Discarding the " + ((CodegenModel)actionModel.get("model")).name + " resource.");
                                actionModel.remove("post_actions");
                                actionModel.remove("operations");
                                break;
                            }
                        }
                        break;
                    }
                }
            }
        }

        // 5.
        for (Object value : models.values()) {
            HashMap<String, Object> model = (HashMap<String, Object>)value;
            String modelUrl = (String)model.get("url");
            if (modelUrl == null) {
                continue;
            }

            String[] urlPieces = modelUrl.split("/");
            String lastParent = null;
            StringBuilder urlSoFar = new StringBuilder();
            for (int i = 0; i < urlPieces.length; i++) {
                if (!urlPieces[i].startsWith("{")) {
                    urlSoFar.append(urlPieces[i]);
                    lastParent = urlSoFar.toString();
                } else {
                    // if any future segments are not keys, then this segment
                    // must have an object qualifier (e.g. {foo[bar]}) and not be
                    // just a field name (e.g. {bar})
                    for (int j = i; j < urlPieces.length; j++) {
                        if (urlPieces[j].startsWith("{") && !urlPieces[j].contains("[")) {
                            CodegenModel parentModel = modelForUrl(lastParent, models);
                            if (parentModel == null) {
                                urlSoFar.append(urlPieces[j]);
                                i = j;
                                break;
                            }
                            String fieldName = urlPieces[j].substring(1, urlPieces[j].length() - 1);
                            String newName = "{" + parentModel.name + "[" + fieldName + "]}";
                            urlSoFar.append(newName);
                            ArrayList<HashMap<String, Object>> keys = (ArrayList<HashMap<String, Object>>)model.get("keys");
                            if (keys != null) {
                                for (HashMap<String, Object> keyEntry : keys) {
                                    String keyName = (String)keyEntry.get("key");
                                    if (keyName.equals(fieldName)) {
                                        keyName = parentModel.name + "." + fieldName;
                                        keyEntry.put("key", keyName);
                                    }
                                }
                                model.put("keys", keys);
                            }
                            i = j;
                            break;
                        } else {
                            urlSoFar.append(urlPieces[j]);
                            if (!urlPieces[j].startsWith("{")) {
                                lastParent = urlSoFar.toString();
                            }
                            i = j;
                        }
                        if (j < urlPieces.length - 1) {
                            urlSoFar.append("/");
                        }
                    }
                }

                if (i < urlPieces.length - 1) {
                    urlSoFar.append("/");
                }
            }

            model.put("url", urlSoFar.toString());
        }
        return models;
    }

    private HashMap<String, Object> findRefModel(CodegenProperty var, Map<String, Object> modelRefs) {
        // find the reference model associated with this property. The property's
        // type should either end with the string "Reference" or have the x-ntap-ref
        // vendor extension set which points to the reference type

        HashMap<String, Object> refMap = null;
        if (var.baseType.endsWith("Reference")) {
            for (String refKey : modelRefs.keySet()) {
                HashMap<String, Object> modelMap = (HashMap<String, Object>)modelRefs.get(refKey);
                CodegenModel refModel = (CodegenModel)modelMap.get("model");

                // if there is no non-reference version of the object, then the
                // classname matching the reference version won't end with Reference
                // we'll just normalize that here so that our check always succeeds
                String classname = refModel.classname;
                if (!classname.endsWith("Reference")) {
                    classname += "Reference";
                }
                if (classname.equals(var.baseType)) {
                    refMap = modelMap;
                    break;
                }
            }
        } else {
            if (var.isListContainer) {
                refMap = (HashMap<String, Object>)modelRefs.get(var.items.vendorExtensions.get("x-ntap-ref"));
            } else {
                refMap = (HashMap<String, Object>)modelRefs.get(var.vendorExtensions.get("x-ntap-ref"));
            }
        }
        return refMap;
    }

    private ArrayList<String> getAccessible(
        CodegenModel model, Map<String, Object> models, Map<String, Object> modelRefs, String[] avoidKeys
    ) {
        ArrayList<String> accessibleVars = new ArrayList<String>();

        // check if there is an x-ntap-readOnly or similiar at the top level of
        // the object, that is outside of the properties list. If so, then consider
        // all of the properties of the object to be non-matching without checking
        // each one
        boolean unwantedKeys = !Collections.disjoint(model.vendorExtensions.keySet(), Arrays.asList(avoidKeys));
        if (unwantedKeys) {
            return accessibleVars;
        }

        for (CodegenProperty var : model.vars) {
            // if the property points to a reference type, then we need to
            // examine the reference's properties to determine which ones are
            // writable for our purpose
            CodegenModel varModel = null;
            HashMap<String, Object> refMap = null;

            for (String modelName : models.keySet()) {
                HashMap<String, Object> varModelMap = (HashMap<String, Object>)models.get(modelName);
                CodegenModel testModel = (CodegenModel)varModelMap.get("model");
                if ((var.isListContainer && var.items != null && testModel.classname.equals(var.items.baseType)) ||
                    testModel.classname.equals(var.baseType)) {
                    varModel = testModel;
                    break;
                }
            }
            if (var.isReadOnly) {
                var.vendorExtensions.put("x-ntap-readOnly", true);
            }
            unwantedKeys = !Collections.disjoint(var.vendorExtensions.keySet(), Arrays.asList(avoidKeys));
            if (varModel != null) {
                unwantedKeys = unwantedKeys || !Collections.disjoint(
                    varModel.vendorExtensions.keySet(), Arrays.asList(avoidKeys)
                );
            }

            // If this var doesn't match our allowed accessor, then just return now
            if (unwantedKeys) {
                continue;
            }

            if (var.isPrimitiveType) {
                // writable primitives go on the list
                accessibleVars.add(var.name);
            } else if (var.isListContainer) {
                // the list is writable, so lets figure out what to write to the list
                if (var.items.isPrimitiveType) {
                    // lists of primitives are good
                    accessibleVars.add(var.name);
                } else if (var.items.vendorExtensions.containsKey("x-ntap-ref") ||
                           var.items.baseType.endsWith("Reference")) {
                    // A list of reference objects so let's figure out which sub
                    // fields are writable
                    refMap = this.findRefModel(var, modelRefs);
                    if (refMap == null) {
                        // its a reference to a non _reference object so just add it
                        // as if it were an embedded object because we will check its
                        // fields when we recurse to it
                        accessibleVars.add(var.name);
                    }
                } else {
                    // if the list doesn't contain a reference type, we don't have
                    // to worry about which sub fields are writable here so it goes
                    // on the list
                    accessibleVars.add(var.name);
                }
            } else if (var.vendorExtensions.containsKey("x-ntap-ref") ||
                       var.baseType.endsWith("Reference")) {
                // This is a reference subobject so we'll figure out which fields
                // are writable
                refMap = this.findRefModel(var, modelRefs);
                if (refMap == null) {
                    // its a reference to a non _reference object so just add it
                    // as if it were an embedded object because we will check its
                    // fields when we recurse to it
                    accessibleVars.add(var.name);
                }
            } else {
                // Finally, this might be an embeded object or a non-inline ref
                // to an object so we'll just add it
                accessibleVars.add(var.name);
            }

            if (refMap != null) {
                CodegenModel referenceModel = (CodegenModel)refMap.get("model");
                for (CodegenProperty refVar : referenceModel.vars) {
                    if (refVar.isReadOnly) {
                        refVar.vendorExtensions.put("x-ntap-readOnly", true);
                    }
                    boolean unwantedRefKeys = !Collections.disjoint(refVar.vendorExtensions.keySet(), Arrays.asList(avoidKeys));
                    if (!unwantedRefKeys) {
                        accessibleVars.add(var.name + "." + refVar.name);
                    }
                }
            }
        }

        return accessibleVars;
    }

    private CodegenModel modelForUrl(String url, Map<String, Object> models) {
        for (Object value : models.values()) {
            HashMap<String, Object> model = (HashMap<String, Object>)value;
            String modelUrl = (String)model.get("url");
            if (modelUrl != null) {
                String dottedStyle = url.replaceAll("\\[", ".").replaceAll("\\]", "");
                if (modelUrl.equals(url) || modelUrl.equals(dottedStyle)) {
                    return (CodegenModel)model.get("model");
                }
            }
        }
        return null;
    }

    private String markdownAdjust(String description) {
        // make any adjustments to the description that make the documentation look better
        ArrayList<String> lines = new ArrayList(Arrays.asList(description.split("\\n")));
        ArrayList<String> newLines = new ArrayList<String>();
        boolean exitedList = false;
        for (int index = 0; index < lines.size(); index++) {
            String originalLine = lines.get(index);
            // There are a bunch of lists in the docs that are using &ndash; instead
            // of a * or a -. This does not work with our markdown parser so we
            // will replace these characters with a * since our logic below is looking
            // for those and many other lists in the docs are made with *s
            originalLine = originalLine.replace("&ndash;", "*");
            if (index == 0) {
                newLines.add(originalLine + "\n");
                continue;
            }
            String trimmedLine = originalLine.trim();
            String trimmedPreviousLine = lines.get(index - 1).trim();
            if (exitedList) {
                exitedList = false;
            } else if (trimmedLine.startsWith("* ") && !trimmedPreviousLine.startsWith("* ") &&
                !trimmedPreviousLine.equals("")) {
                // here we add a blank line just before a list of items if there
                // wasn't already one. Strict markdown renderers require this.
                // Example - this:
                //
                // This is a list of things:
                // * foo
                // * bar
                //
                // becomes this:
                //
                // This is a list of things:
                //
                // * foo
                // * bar
                newLines.add("\n");
            }

            String lineToAdd = originalLine;
            if (index < lines.size() - 1 && trimmedLine.startsWith("* ")) {
                while (true) {
                    // In this loop, we are turning multiline list items into a
                    // single line. Needed for using pdoc to render the documentation.
                    // Example - this:
                    //
                    // * this is a multiline
                    //   item which should be one line
                    // * this is another
                    //   item
                    //
                    // becomes this:
                    //
                    // * this is a multiline item which should be one line
                    // * this is another item
                    String nextLine = lines.get(index + 1);
                    int numSpaces = originalLine.indexOf('*');
                    if (ltrim(nextLine).length() + numSpaces + 2 == nextLine.length()) {
                        lineToAdd += " " + nextLine.trim();
                        index += 1;
                    } else {
                        exitedList = true;
                        lineToAdd += "\n";
                        break;
                    }
                }
            } else if (index < lines.size() - 1) {
                lineToAdd += "\n";
            }

            newLines.add(lineToAdd);
        }

        StringBuilder newDescription = new StringBuilder();
        for (String line : newLines) {
            newDescription.append(line);
        }

        return this.rtrim(newDescription.toString());
    }

    private String ltrim(String s) {
        int i = 0;
        while (i < s.length() && Character.isWhitespace(s.charAt(i))) {
            i++;
        }
        return s.substring(i);
    }

    private String rtrim(String s) {
        int i = s.length() - 1;
        while(i >= 0 && Character.isWhitespace(s.charAt(i))) {
            i--;
        }
        return s.substring(0, i + 1);
    }

    private boolean isBulkPostSupported(Map<String, Object> models, String modelName, io.swagger.models.Path path) {
        // bulk post is supported if the model has a "records" property or
        // x-ntap-allow-bulk-post is set at the top level of the YAML
        // bulk post will be disallowed if x-ntap-allow-bulk-post is set to false
        // on the path level (this will override both of the prior checks)
        String bulk_post_extension = "x-ntap-allow-bulk-post";
        boolean modelSupported = hasRecords(models, modelName);
        boolean globalSupported = false;
        if (swagger.getVendorExtensions() != null &&
            swagger.getVendorExtensions().containsKey(bulk_post_extension)) {
            globalSupported = (Boolean)swagger.getVendorExtensions().get(bulk_post_extension);
        }
        boolean pathSupported = true;
        if (path.getVendorExtensions() != null &&
            path.getVendorExtensions().containsKey(bulk_post_extension)) {
            pathSupported = (Boolean)path.getVendorExtensions().get(bulk_post_extension);
        }

        if (!pathSupported) {
            return false;
        }
        return modelSupported || globalSupported;
    }

    private boolean hasRecords(Map<String, Object> models, String modelName) {
        boolean hasRecordsVar = false;
        HashMap<String, Object> mySchema = (HashMap<String, Object>)models.get(modelName);
        CodegenModel cgModel = (CodegenModel)mySchema.get("model");
        for (CodegenProperty var : cgModel.vars) {
            if(var.name.equals("records")) {
                hasRecordsVar = true;
                break;
            }
        }

        return hasRecordsVar;
    }

    private boolean hasRecordsAndNumRecords(CodegenModel cgModel) {
        boolean hasRecords = false;
        boolean hasNumRecords = false;
        for (CodegenProperty var : cgModel.vars) {
            if(var.name.equals("records")) {
                hasRecords = true;
            }
            if(var.name.equals("num_records")) {
                hasNumRecords = true;
            }
        }

        return hasRecords && hasNumRecords;
    }

    protected void generateResources(List<File> files, List<Object> allResources, List<Object> allResourceModels) {
        if (!isGenerateResources) {
            return;
        }

        // Put the two processed pieces together for the template engine
        Map<String, Object> resources = linkPathsAndModels();

        // Only generate the resources we are told to
        Set<String> resourcesToGenerate = null;
        String resourceNames = System.getProperty("resources");
        if(resourceNames != null && !resourceNames.isEmpty()) {
            resourcesToGenerate = new HashSet<String>(Arrays.asList(resourceNames.split(",")));
        }
        if(resourcesToGenerate != null && !resourcesToGenerate.isEmpty()) {
            Map<String, Object> updatedResources = new HashMap<>();
            for(String m : resources.keySet()) {
                if(resourcesToGenerate.contains(m)) {
                    updatedResources.put(m, resources.get(m));
                }
            }
            resources = updatedResources;
        }

        // we are going to track for each complex type if it has
        // operations or not because they will be put into different
        // packages (models and resources). First, look at each resource we've
        // generated and decide if it is a resource or model (has operations or not)
        for (String resourceName : resources.keySet()) {
            String packageName = "models";
            HashMap<String, Object> resourceMap = (HashMap<String, Object>)resources.get(resourceName);
            if (resourceMap.containsKey("operations") || resourceMap.containsKey("post_actions")) {
                packageName = "resources";
            }
            resourceMap.put("packageName", packageName);
        }

        // then for each resoruce we've generated, we need to mark each of its
        // fields as either a model or resource based on the information we just
        // marked for that type
        for (String resourceName : resources.keySet()) {
            HashMap<String, Object> schema = (HashMap<String, Object>)resources.get(resourceName);
            CodegenModel model = (CodegenModel)schema.get("model");
            for (CodegenProperty var : model.vars) {
                if (var.complexType == null) {
                    continue;
                }
                String module = config.toResourceFilename(var.baseType);
                if (var.isListContainer) {
                    if (var.items != null) {
                        module = config.toResourceFilename(var.items.baseType);
                    } else {
                        module = config.toResourceFilename(var.baseType);
                    }
                }
                if (module.endsWith("_reference")) {
                    module = module.substring(0, module.length() - 10);
                }
                HashMap<String, Object> varSchema = (HashMap<String, Object>)resources.get(module);
                if (varSchema == null) {
                    var.vendorExtensions.put("packageName", "models." + module);
                } else {
                    var.vendorExtensions.put("packageName", varSchema.get("packageName") + "." + module);
                }
            }
        }

        // Generate resource files
        for (String resourceName : resources.keySet()) {
            try {
                HashMap<String, Object> resourceTemplate = new HashMap<String, Object>();
                resourceTemplate.put("year", Calendar.getInstance().get(Calendar.YEAR));
                resourceTemplate.put("resource", (HashMap<String, Object>) resources.get(resourceName));
                if (config instanceof OntapPythonClientCodegen) {
                    resourceTemplate.put("pythonPackageName", config.additionalProperties().get(CodegenConstants.PACKAGE_NAME));
                }
                for (String templateName : config.resourceTemplateFiles().keySet()) {
                    HashMap<String, Object> resourceMap = (HashMap<String, Object>)resources.get(resourceName);
                    if (resourceMap.containsKey("operations") || resourceMap.containsKey("post_actions")) {
                        allResources.add(resourceTemplate);
                        String suffix = config.resourceTemplateFiles().get(templateName);
                        String filename = config.resourceFileFolder() + File.separator + config.toResourceFilename(resourceName) + suffix;
                        if (!config.shouldOverwrite(filename)) {
                            //LOGGER.info("Skipped overwriting " + filename);
                            continue;
                        }
                        File written = processTemplateToFile(resourceTemplate, templateName, filename);
                        if(written != null) {
                            files.add(written);
                        }
                    } else {
                        allResourceModels.add(resourceTemplate);
                        String suffix = config.resourceTemplateFiles().get(templateName);
                        String filename = config.modelFileFolder() + File.separator + config.toModelFilename(resourceName) + suffix;
                        if (!config.shouldOverwrite(filename)) {
                            //LOGGER.info("Skipped overwriting " + filename);
                            continue;
                        }
                        File written = processTemplateToFile(resourceTemplate, "model_schema.mustache", filename);
                        if(written != null) {
                            files.add(written);
                        }
                    }
                }
            } catch (Exception e) {
                throw new RuntimeException("Could not genereate resource '" + resourceName + "'", e);
            }
        }
        if (allResources.size() > 0) {
            HashMap<String, Object> lastResource = (HashMap<String, Object>)allResources.get(allResources.size() - 1);
            lastResource.put("lastResource", true);
        }
        if (allResourceModels.size() > 0) {
            HashMap<String, Object> lastResourceModel = (HashMap<String, Object>)allResourceModels.get(allResourceModels.size() - 1);
            lastResourceModel.put("lastResourceModel", true);
        }

        if (System.getProperty("debugResources") != null) {
            //LOGGER.info("############ Resource info ############");
            Json.prettyPrint(allResources);
            Json.prettyPrint(allResourceModels);
        }
    }

    protected void generateSupportingFiles(List<File> files, Map<String, Object> bundle, List<String> br_gen) {
        if (!isGenerateSupportingFiles) {
            return;
        }
        Set<String> supportingFilesToGenerate = null;
        String supportingFiles = System.getProperty(CodegenConstants.SUPPORTING_FILES);
        if (supportingFiles != null && !supportingFiles.isEmpty()) {
            supportingFilesToGenerate = new HashSet<String>(Arrays.asList(supportingFiles.split(",")));
        }

        for (SupportingFile support : config.supportingFiles()) {
            try {
                String outputFolder = config.outputFolder();
                if (StringUtils.isNotEmpty(support.folder)) {
                    outputFolder += File.separator + support.folder;
                }
                File of = new File(outputFolder);
                if (!of.isDirectory()) {
                    of.mkdirs();
                }
                String outputFilename = outputFolder + File.separator + support.destinationFilename.replace('/', File.separatorChar);
                if (support.destinationFilename.endsWith(".cpp")) {
                    String br_supp_path = this.srcDirPrefix + File.separator;
                    if (StringUtils.isNotEmpty(support.folder)) {
                        br_supp_path += support.folder + File.separator;
                    }
                    br_supp_path += support.destinationFilename;
                    br_gen.add(br_supp_path);
                }
                if (!config.shouldOverwrite(outputFilename)) {
                    //LOGGER.info("Skipped overwriting " + outputFilename);
                    continue;
                }
                String templateFile;
                if (support instanceof GlobalSupportingFile) {
                    templateFile = config.getCommonTemplateDir() + File.separator + support.templateFile;
                } else {
                    templateFile = getFullTemplateFile(config, support.templateFile);
                }
                boolean shouldGenerate = true;
                if (supportingFilesToGenerate != null && !supportingFilesToGenerate.isEmpty()) {
                    shouldGenerate = supportingFilesToGenerate.contains(support.destinationFilename);
                }
                if (!shouldGenerate) {
                    continue;
                }

                if (ignoreProcessor.allowsFile(new File(outputFilename))) {
                    if (templateFile.endsWith("mustache")) {
                        String template = readTemplate(templateFile);
                        Mustache.Compiler compiler = Mustache.compiler();
                        compiler = config.processCompiler(compiler);
                        Template tmpl = compiler
                                .withLoader(new Mustache.TemplateLoader() {
                                    @Override
                                    public Reader getTemplate(String name) {
                                        return getTemplateReader(getFullTemplateFile(config, name + ".mustache"));
                                    }
                                })
                                .defaultValue("")
                                .compile(template);

                        writeToFile(outputFilename, tmpl.execute(bundle));
                        files.add(new File(outputFilename));
                    } else {
                        InputStream in = null;

                        try {
                            in = new FileInputStream(templateFile);
                        } catch (Exception e) {
                            // continue
                        }
                        if (in == null) {
                            in = this.getClass().getClassLoader().getResourceAsStream(getCPResourcePath(templateFile));
                        }
                        File outputFile = new File(outputFilename);
                        OutputStream out = new FileOutputStream(outputFile, false);
                        if (in != null) {
                            //LOGGER.info("writing file " + outputFile);
                            IOUtils.copy(in, out);
                            out.close();
                        } else {
                            LOGGER.error("can't open " + templateFile + " for input");
                        }
                        files.add(outputFile);
                    }
                } else {
                    //LOGGER.info("Skipped generation of " + outputFilename + " due to rule in .swagger-codegen-ignore");
                }
            } catch (Exception e) {
                throw new RuntimeException("Could not generate supporting file '" + support + "'", e);
            }
        }

        // Consider .swagger-codegen-ignore a supporting file
        // Output .swagger-codegen-ignore if it doesn't exist and wasn't explicitly created by a generator
        final String swaggerCodegenIgnore = ".swagger-codegen-ignore";
        String ignoreFileNameTarget = config.outputFolder() + File.separator + swaggerCodegenIgnore;
        File ignoreFile = new File(ignoreFileNameTarget);
        if (isGenerateSwaggerMetadata && !ignoreFile.exists()) {
            String ignoreFileNameSource = File.separator + config.getCommonTemplateDir() + File.separator + swaggerCodegenIgnore;
            String ignoreFileContents = readResourceContents(ignoreFileNameSource);
            try {
                writeToFile(ignoreFileNameTarget, ignoreFileContents);
            } catch (IOException e) {
                throw new RuntimeException("Could not generate supporting file '" + swaggerCodegenIgnore + "'", e);
            }
            files.add(ignoreFile);
        }

        if(isGenerateSwaggerMetadata) {
            final String swaggerVersionMetadata = config.outputFolder() + File.separator + ".swagger-codegen" + File.separator + "VERSION";
            File swaggerVersionMetadataFile = new File(swaggerVersionMetadata);
            try {
                writeToFile(swaggerVersionMetadata, ImplementationVersion.read());
                files.add(swaggerVersionMetadataFile);
            } catch (IOException e) {
                throw new RuntimeException("Could not generate supporting file '" + swaggerVersionMetadata + "'", e);
            }
        }

        /*
         * The following code adds default LICENSE (Apache-2.0) for all generators
         * To use license other than Apache2.0, update the following file:
         *   modules/swagger-codegen/src/main/resources/_common/LICENSE
         *
        final String apache2License = "LICENSE";
        String licenseFileNameTarget = config.outputFolder() + File.separator + apache2License;
        File licenseFile = new File(licenseFileNameTarget);
        String licenseFileNameSource = File.separator + config.getCommonTemplateDir() + File.separator + apache2License;
        String licenseFileContents = readResourceContents(licenseFileNameSource);
        try {
            writeToFile(licenseFileNameTarget, licenseFileContents);
        } catch (IOException e) {
            throw new RuntimeException("Could not generate LICENSE file '" + apache2License + "'", e);
        }
        files.add(licenseFile);
         */

    }

    protected Map<String, Object> buildSupportFileBundle(List<Object> allOperations, List<Object> allModels, List<Object> allResources, List<Object> allResourceModels) {

        Map<String, Object> bundle = new HashMap<String, Object>();
        bundle.putAll(config.additionalProperties());
        bundle.put("apiPackage", config.apiPackage());

        Map<String, Object> apis = new HashMap<String, Object>();
        apis.put("apis", allOperations);

        if (swagger.getHost() != null) {
            bundle.put("host", swagger.getHost());
        }

        bundle.put("swagger", this.swagger);
        bundle.put("basePath", basePath);
        bundle.put("basePathWithoutHost", basePathWithoutHost);
        bundle.put("scheme", getScheme());
        bundle.put("contextPath", contextPath);
        bundle.put("apiInfo", apis);
        bundle.put("resources", allResources);
        bundle.put("resourceModels", allResourceModels);
        bundle.put("models", allModels);
        bundle.put("apiFolder", config.apiPackage().replace('.', File.separatorChar));
        bundle.put("modelPackage", config.modelPackage());
        if (config instanceof OntapPythonClientCodegen) {
            bundle.put("pythonPackageName", config.additionalProperties().get(CodegenConstants.PACKAGE_NAME));
        }
        List<CodegenSecurity> authMethods = config.fromSecurity(swagger.getSecurityDefinitions());
        if (authMethods != null && !authMethods.isEmpty()) {
            bundle.put("authMethods", authMethods);
            bundle.put("hasAuthMethods", true);
        }
        if (swagger.getExternalDocs() != null) {
            bundle.put("externalDocs", swagger.getExternalDocs());
        }
        for (int i = 0; i < allModels.size() - 1; i++) {
            HashMap<String, CodegenModel> cm = (HashMap<String, CodegenModel>) allModels.get(i);
            CodegenModel m = cm.get("model");
            m.hasMoreModels = true;
        }

        config.postProcessSupportingFileData(bundle);

        if (System.getProperty("debugSupportingFiles") != null) {
            //LOGGER.info("############ Supporting file info ############");
            Json.prettyPrint(bundle);
        }
        return bundle;
    }

    @Override
    public List<File> generate() {

        if (swagger == null || config == null) {
            throw new RuntimeException("missing swagger input or config!");
        }
        configureGeneratorProperties();
        configureSwaggerInfo();

        // resolve inline models
        InlineModelResolver inlineModelResolver = new InlineModelResolver();
        inlineModelResolver.flatten(swagger);

        List<String> br_gen = new ArrayList<String>();
        List<File> files = new ArrayList<File>();
        // models
        List<Object> allModels = new ArrayList<Object>();
        generateModels(files, allModels, br_gen);
        // apis
        List<Object> allOperations = new ArrayList<Object>();
        generateApis(files, allOperations, allModels, br_gen);
        // resources
        List<Object> allResources = new ArrayList<>();
        List<Object> allResourceModels = new ArrayList<>();
        generateResources(files, allResources, allResourceModels);

        // supporting files
        Map<String, Object> bundle = buildSupportFileBundle(allOperations, allModels, allResources, allResourceModels);
        generateSupportingFiles(files, bundle, br_gen);
        config.processSwagger(swagger);

// NETAPP
        if (this.srcDirPrefix != null) {
            String br_gen_path = config.outputFolder() + File.separator + ".BR_SWAGGER_";
            if (config instanceof PistacheServerCodegen) {
                br_gen_path += "PISTACHE_SERVER";
            } else {
                br_gen_path += this.srcDirPrefix.substring(this.srcDirPrefix.lastIndexOf("/") + 1).toUpperCase();
            }
            br_gen_path += "_CC";
            LOGGER.warn("BRP - Using br_gen_path 2 {}", br_gen_path);
            java.nio.file.Path file = java.nio.file.Paths.get(br_gen_path);
            try {
                Files.write(file, br_gen, StandardCharsets.UTF_8);
            } catch (Exception e) {
                throw new RuntimeException("Could not generate .BR_SWAGGER_SERVER_CC file", e);
            }
        }
// end NETAPP
        return files;
    }


    protected File processTemplateToFile(Map<String, Object> templateData, String templateName, String outputFilename) throws IOException {
        String adjustedOutputFilename = outputFilename.replaceAll("//", "/").replace('/', File.separatorChar);
        if (ignoreProcessor.allowsFile(new File(adjustedOutputFilename))) {
            String templateFile = getFullTemplateFile(config, templateName);
            String template = readTemplate(templateFile);
            Mustache.Compiler compiler = Mustache.compiler();
            compiler = config.processCompiler(compiler);
            Template tmpl = compiler
                    .withLoader(new Mustache.TemplateLoader() {
                        @Override
                        public Reader getTemplate(String name) {
                            return getTemplateReader(getFullTemplateFile(config, name + ".mustache"));
                        }
                    })
                    .defaultValue("")
                    .compile(template);

            writeToFile(adjustedOutputFilename, tmpl.execute(templateData));
            return new File(adjustedOutputFilename);
        }

        //LOGGER.info("Skipped generation of " + adjustedOutputFilename + " due to rule in .swagger-codegen-ignore");
        return null;
    }

    protected static void processMimeTypes(List<String> mimeTypeList, Map<String, Object> operation, String source) {
        if (mimeTypeList == null || mimeTypeList.isEmpty()) {
            return;
        }
        List<Map<String, String>> c = new ArrayList<Map<String, String>>();
        int count = 0;
        for (String key : mimeTypeList) {
            Map<String, String> mediaType = new HashMap<String, String>();
            mediaType.put("mediaType", key);
            count += 1;
            if (count < mimeTypeList.size()) {
                mediaType.put("hasMore", "true");
            } else {
                mediaType.put("hasMore", null);
            }
            c.add(mediaType);
        }
        operation.put(source, c);
        String flagFieldName = "has" + source.substring(0, 1).toUpperCase() + source.substring(1);
        operation.put(flagFieldName, true);

    }

    public Map<String, List<CodegenOperation>> processPaths(Map<String, io.swagger.models.Path> paths) {
        Map<String, List<CodegenOperation>> ops = new TreeMap<String, List<CodegenOperation>>();
        for (String resourcePath : paths.keySet()) {
            io.swagger.models.Path path = paths.get(resourcePath);
            processOperation(resourcePath, "get", path.getGet(), ops, path);
            processOperation(resourcePath, "head", path.getHead(), ops, path);
            processOperation(resourcePath, "put", path.getPut(), ops, path);
            processOperation(resourcePath, "post", path.getPost(), ops, path);
            processOperation(resourcePath, "delete", path.getDelete(), ops, path);
            processOperation(resourcePath, "patch", path.getPatch(), ops, path);
            processOperation(resourcePath, "options", path.getOptions(), ops, path);
        }
        return ops;
    }

    protected void processOperation(String resourcePath, String httpMethod, Operation operation, Map<String, List<CodegenOperation>> operations, io.swagger.models.Path path) {
        if (operation == null) {
            return;
        }
        if (System.getProperty("debugOperations") != null) {
            //LOGGER.info("processOperation: resourcePath= " + resourcePath + "\t;" + httpMethod + " " + operation + "\n");
        }
        List<Tag> tags = new ArrayList<Tag>();

        List<String> tagNames = operation.getTags();
        List<Tag> swaggerTags = swagger.getTags();
        if (tagNames != null) {
            if (swaggerTags == null) {
                for (String tagName : tagNames) {
                    tags.add(new Tag().name(tagName));
                }
            } else {
                for (String tagName : tagNames) {
                    boolean foundTag = false;
                    for (Tag tag : swaggerTags) {
                        if (tag.getName().equals(tagName)) {
                            tags.add(tag);
                            foundTag = true;
                            break;
                        }
                    }

                    if (!foundTag) {
                        tags.add(new Tag().name(tagName));
                    }
                }
            }
        }

        if (tags.isEmpty()) {
            tags.add(new Tag().name("default"));
        }

        /*
         build up a set of parameter "ids" defined at the operation level
         per the swagger 2.0 spec "A unique parameter is defined by a combination of a name and location"
          i'm assuming "location" == "in"
        */
        Set<String> operationParameters = new HashSet<String>();
        if (operation.getParameters() != null) {
            for (Parameter parameter : operation.getParameters()) {
                operationParameters.add(generateParameterId(parameter));
            }
        }

        //need to propagate path level down to the operation
        if (path.getParameters() != null) {
            for (Parameter parameter : path.getParameters()) {
                //skip propagation if a parameter with the same name is already defined at the operation level
                if (!operationParameters.contains(generateParameterId(parameter))) {
                    operation.addParameter(parameter);
                }
            }
        }

        for (Tag tag : tags) {
            try {
                CodegenOperation codegenOperation = config.fromOperation(resourcePath, httpMethod, operation, swagger.getDefinitions(), swagger);
                codegenOperation.tags = new ArrayList<Tag>(tags);
                config.addOperationToGroup(config.sanitizeTag(tag.getName()), resourcePath, operation, codegenOperation, operations);

                List<Map<String, List<String>>> securities = operation.getSecurity();
                if (securities == null && swagger.getSecurity() != null) {
                    securities = new ArrayList<Map<String, List<String>>>();
                    for (SecurityRequirement sr : swagger.getSecurity()) {
                        securities.add(sr.getRequirements());
                    }
                }
                if (securities == null || swagger.getSecurityDefinitions() == null) {
                    continue;
                }
                Map<String, SecuritySchemeDefinition> authMethods = new HashMap<String, SecuritySchemeDefinition>();
                for (Map<String, List<String>> security : securities) {
                    for (String securityName : security.keySet()) {
                        SecuritySchemeDefinition securityDefinition = swagger.getSecurityDefinitions().get(securityName);
                        if (securityDefinition == null) {
                            continue;
                        }
                        if (securityDefinition instanceof OAuth2Definition) {
                            OAuth2Definition oauth2Definition = (OAuth2Definition) securityDefinition;
                            OAuth2Definition oauth2Operation = new OAuth2Definition();
                            oauth2Operation.setType(oauth2Definition.getType());
                            oauth2Operation.setAuthorizationUrl(oauth2Definition.getAuthorizationUrl());
                            oauth2Operation.setFlow(oauth2Definition.getFlow());
                            oauth2Operation.setTokenUrl(oauth2Definition.getTokenUrl());
                            oauth2Operation.setScopes(new HashMap<String, String>());
                            for (String scope : security.get(securityName)) {
                                if (oauth2Definition.getScopes().containsKey(scope)) {
                                    oauth2Operation.addScope(scope, oauth2Definition.getScopes().get(scope));
                                }
                            }
                            authMethods.put(securityName, oauth2Operation);
                        } else {
                            authMethods.put(securityName, securityDefinition);
                        }
                    }
                }
                if (!authMethods.isEmpty()) {
                    codegenOperation.authMethods = config.fromSecurity(authMethods);
                    codegenOperation.hasAuthMethods = true;
                }
            } catch (Exception ex) {
                String msg = "Could not process operation:\n" //
                        + "  Tag: " + tag + "\n"//
                        + "  Operation: " + operation.getOperationId() + "\n" //
                        + "  Resource: " + httpMethod + " " + resourcePath + "\n"//
                        + "  Definitions: " + swagger.getDefinitions() + "\n"  //
                        + "  Exception: " + ex.getMessage();
                throw new RuntimeException(msg, ex);
            }
        }

    }

    protected static String generateParameterId(Parameter parameter) {
        return parameter.getName() + ":" + parameter.getIn();
    }


    protected Map<String, Object> processOperations(CodegenConfig config, String tag, List<CodegenOperation> ops, List<Object> allModels) {
        Map<String, Object> operations = new HashMap<String, Object>();
        Map<String, Object> objs = new HashMap<String, Object>();
        objs.put("classname", config.toApiName(tag));
        objs.put("pathPrefix", config.toApiVarName(tag));

        // check for operationId uniqueness
        Set<String> opIds = new HashSet<String>();
        int counter = 0;
        for (CodegenOperation op : ops) {
            String opId = op.nickname;
            if (opIds.contains(opId)) {
                counter++;
                op.nickname += "_" + counter;
            }
            opIds.add(opId);
        }
        objs.put("operation", ops);

        operations.put("operations", objs);
        operations.put("package", config.apiPackage());


        Set<String> allImports = new TreeSet<String>();
        for (CodegenOperation op : ops) {
            allImports.addAll(op.imports);
        }

        List<Map<String, String>> imports = new ArrayList<Map<String, String>>();
        for (String nextImport : allImports) {
            Map<String, String> im = new LinkedHashMap<String, String>();
            String mapping = config.importMapping().get(nextImport);
            if (mapping == null) {
                mapping = config.toModelImport(nextImport);
            }
            if (mapping != null) {
                im.put("import", mapping);
                if (!imports.contains(im)) { // avoid duplicates
                    imports.add(im);
                }
            }
        }

        operations.put("imports", imports);

        // add a flag to indicate whether there's any {{import}}
        if (imports.size() > 0) {
            operations.put("hasImport", true);
        }
        config.postProcessOperations(operations);
        config.postProcessOperationsWithModels(operations, allModels);
        if (objs.size() > 0) {
            List<CodegenOperation> os = (List<CodegenOperation>) objs.get("operation");

            if (os != null && os.size() > 0) {
                CodegenOperation op = os.get(os.size() - 1);
                op.hasMore = false;
            }
        }
        return operations;
    }

    protected Map<String, Object> processModels(CodegenConfig config, Map<String, Model> definitions, Map<String, Model> allDefinitions) {
        Map<String, Object> objs = new HashMap<String, Object>();
        objs.put("package", config.modelPackage());
        List<Object> models = new ArrayList<Object>();
        Set<String> allImports = new LinkedHashSet<String>();
        for (String key : definitions.keySet()) {
            Model mm = definitions.get(key);
            if(mm.getVendorExtensions() !=  null && mm.getVendorExtensions().containsKey("x-codegen-ignore")) {
                // skip this model
                LOGGER.debug("skipping model " + key);
                return null;
            }
            else if(mm.getVendorExtensions() !=  null && mm.getVendorExtensions().containsKey("x-codegen-import-mapping")) {
                String codegenImport = mm.getVendorExtensions().get("x-codegen-import-mapping").toString();
                config.importMapping().put(key, codegenImport);
                allImports.add(codegenImport);
            }
            CodegenModel cm = config.fromModel(key, mm, allDefinitions);
            Map<String, Object> mo = new HashMap<String, Object>();
            mo.put("model", cm);
            mo.put("importPath", config.toModelImport(cm.classname));
            models.add(mo);

            allImports.addAll(cm.imports);
        }
        objs.put("models", models);
        Set<String> importSet = new TreeSet<String>();
        for (String nextImport : allImports) {
            String mapping = config.importMapping().get(nextImport);
            if (mapping == null) {
                mapping = config.toModelImport(nextImport);
            }
            if (mapping != null && !config.defaultIncludes().contains(mapping)) {
                importSet.add(mapping);
            }
            // add instantiation types
            mapping = config.instantiationTypes().get(nextImport);
            if (mapping != null && !config.defaultIncludes().contains(mapping)) {
                importSet.add(mapping);
            }
        }
        List<Map<String, String>> imports = new ArrayList<Map<String, String>>();
        for (String s : importSet) {
            Map<String, String> item = new HashMap<String, String>();
            item.put("import", s);
            imports.add(item);
        }
        objs.put("imports", imports);
        config.postProcessModels(objs);
        return objs;
    }
}
